table.stats {
  width: 100%;
  border-radius: 3px;
  border: 2px solid #11251c;
  border-spacing: 0;
  margin-bottom: 30px; }
  table.stats thead {
    padding: 0;
    margin: 0;
    color: #a9c5ba; }
    table.stats thead tr {
      background: #11251c; }
      table.stats thead tr:nth-child(even) {
        background: #11251c; }
      table.stats thead tr:hover {
        background: #11251c; }
    table.stats thead th {
      font-weight: bold;
      cursor: default; }
      table.stats thead th.nowrap {
        white-space: nowrap; }
      table.stats thead th[data-sortkey] {
        cursor: pointer; }
  table.stats tr:hover {
    background: #1d4434; }
  table.stats tr td, table.stats tr th {
    padding: 10px;
    margin: 0;
    text-align: left; }
    table.stats tr td:first-child, table.stats tr th:first-child {
      padding-left: 16px; }
    table.stats tr td:last-child, table.stats tr th:last-child {
      padding-right: 16px; }
    table.stats tr td.numeric, table.stats tr th.numeric {
      text-align: right; }
  table.stats tr td {
    max-width: 450px;
    overflow: hidden;
    word-wrap: break-word;
    white-space: no-wrap;
    text-overflow: ellipsis; }
    @media (max-width: 1100px) {
      table.stats tr td {
        max-width: 350px; } }
  table.stats tr:nth-child(even) {
    background: #153126; }
    table.stats tr:nth-child(even):hover {
      background: #1d4434; }
  table.stats tr.total {
    background: #11251c; }
    table.stats tr.total td {
      font-weight: bold; }

/*# sourceMappingURL=tables.css.map */
